import ast
import keyboard
import platform
from ._scp import SCPCommand
from ._exec import Execute
from runtime import Runtime
if platform.system() == "Windows":
    import pythonping
else:
    import os


def exec(command: str) -> Execute:
    '''
    Execute command.

    Args:
        command (str): system command or running app.

    Examples:
        >>> exec("hello_world.exe")
    '''
    assert isinstance(command, str)
    _exec = Execute()
    _exec(command)
    return _exec


def scp(command: str) -> None:
    '''
    Secure copy

    Args:
        command (str): scp [-r] [-P port] [[user@]host1:]file1 ... [[user@]host2:]file2 password

    Examples:
        >>> scp("-r -P 23 root@169.254.1.32:/root/a.txt d:\\")
    '''
    assert isinstance(command, str)

    _scp = SCPCommand()
    _scp(command)


def calc(algo):
    '''
    Calculation.

    Args:
        algo: algorithm formula.

    Returns:
        int/float, caculate result.

    Examples:
        >>> print(calc(1+2*5))
    '''
    assert isinstance(algo, (float, int))

    return algo


def input(msg):
    '''
    Input value from UI.

    Args:
        msg(str): ui message.

    Returns:
        Any, inpurt value.

    Examples:
        >>> print(input('values'))
    '''
    data = Runtime.open_ui('input', Runtime.get_channel(), msg)

    # transfor to python data type.
    try:
        return ast.literal_eval(data)
    except Exception:
        return data


def output(command):
    '''
    Input value to UI.

    Args:
        command(str): ui message.

    Examples:
        >>> print(output('ok'))
    '''
    return Runtime.open_ui('dialogbox', Runtime.get_channel(), command)


def combobox_input(msg, items):
    '''
    Select value from combobox UI.

    Args:
        msg: str, ui message.
        items: list, value list.

    Returns:
        Any, inpurt value.

    Examples:
        >>> print(input('values'))
    '''
    data = Runtime.open_ui('combobox_input', Runtime.get_channel(), msg, items)

    # transfor to python data type.
    try:
        return ast.literal_eval(data)
    except Exception:
        return data


def confirm_output(msg):
    '''
    Output value to UI.

    Args:
        msg: str, ui message.

    Examples:
        >>> print(confirm_output('ok'))
    '''
    return Runtime.open_ui('confirm_output', Runtime.get_channel(), msg)


def confirm_stop_output(msg):
    '''
    Output message to UI.

    Args:
        msg: str, ui message.

    Examples:
        >>> print(confirm_stop_output('ok'))
    '''
    return Runtime.open_ui('confirm_stop_output', Runtime.get_channel(), msg)


def alert_output(msg):
    '''
    Output message to UI.

    Args:
        msg: str, ui message.

    Examples:
        >>> print(alert_output('ok'))
    '''
    return Runtime.open_ui('alert_output', Runtime.get_channel(), msg)


def error_output(msg):
    '''
    Output message to UI.

    Args:
        msg: str, ui message.

    Examples:
        >>> print(error_output('ok'))
    '''
    return Runtime.open_ui('error_output', Runtime.get_channel(), msg)


def scaner(mode='hide', length: int = None) -> str:
    '''
    Scaner barcode.

    Args:
        length: int, barcode length.
        mode: str, ['hide', 'show'], when 'show' is set, show the scaner UI.

    Returns:
        str, barcode.

    Examples:
        >>> print(scaner())
    '''
    assert mode in ['hide', 'show']
    if length is not None:
        assert isinstance(length, int)

    if mode == 'show':
        # Sync open UI.
        return Runtime.open_ui('scaner')

    barcode = ''
    Runtime.action('Plugins scaner open')
    # clear keyboard evebr
    keyboard.unhook_all()
    while True:
        # Only press down event
        event = keyboard.read_event(False)
        if event.event_type == keyboard.KEY_UP:
            continue

        key = event.name or event.scan_code
        if not length and key == 'enter':
            break

        if len(key) != 1:
            continue

        barcode += key
        if length and len(barcode) >= length:
            break

    return barcode


def set_sn(serial_number: str):
    '''
    Set serial number.

    Args:
        serial_number (str): serial number

    Returns:
        str, serial number.

    Examples:
        >>> print(set_sn('GQQ02950500Q17584'))
    '''
    return serial_number


def ping(command: str):
    '''
    Pings a remote host and handles the responses

    Args:
        command (str): The remote hostname or IP address to ping,
            ping [options] <destination>.

            Options:
                <destination>      dns name or ip address
                -c <count>         stop after <count> replies
                -W <timeout>       time to wait for response
                -i <interval>      seconds between sending each packet
                -s <size>          use <size> as number of data bytes to be sent

    Examples:
        >>> print(ping('-c 1 -W 0.1 169.254.1.32'))
    '''
    assert isinstance(command, str)

    args = command.split()
    if not args:
        raise ValueError("usage error: Destination address required")

    if platform.system() == "Darwin":
        ping_cmd = os.popen("ping {}".format(command), buffering=1)
        result = b''
        while True:
            data = os.read(ping_cmd.fileno(), 1024)
            result += data
            if not data:
                break
        return result.decode('utf-8')

    timeout = 2
    count = 4
    size = 1
    interval = 0
    target = args[-1]
    flag = None
    arg_len = len(args)
    # decode args.
    for pos in range(arg_len):
        arg = args[pos]
        if '-' in arg and arg not in ['-c', '-W', '-i', '-s']:
            if len(arg) > 2 and arg[1] in ['c', 'W', 'i', 's']:
                raise ValueError("invalid argument: '{}'".format(arg[2]))

            raise ValueError('''ping: invalid option -- '{}'

            Usage
            ping [options] <destination>
            Options:
                <destination>      dns name or ip address
                -c <count>         stop after <count> replies
                -W <timeout>       time to wait for response
                -i <interval>      seconds between sending each packet
                -s <size>          use <size> as number of data bytes to be sent'''.format(arg[1]))

        if flag:
            if pos == arg_len - 1:
                raise ValueError("usage error: Destination address required")

            if flag == '-c':
                count = int(arg)
            elif flag == '-W':
                timeout = float(arg)
            elif flag == '-i':
                interval = int(arg)
            else:
                size = int(arg)
            flag = None
            continue

        if '-' not in arg:
            continue

        flag = arg

    if flag:
        raise ValueError('''option requires an argument -- '{}'

            Usage
            ping [options] <destination>
            Options:
                <destination>      dns name or ip address
                -c <count>         stop after <count> replies
                -W <timeout>       time to wait for response
                -i <interval>      seconds between sending each packet
                -s <size>          use <size> as number of data bytes to be sent'''.format(flag[1]))

    return str(pythonping.ping(target, timeout, count, size, interval))
