from typing import Tuple
from decimal import Context
from decimal import ROUND_HALF_UP
from decimal import ROUND_UP
from decimal import ROUND_DOWN


def string_to_hex(src: str) -> list:
    '''Transfer string data to hex list.

    Args:
        src (str): source string.

    Returns:
        list: 8bit int data list.

    Examples:
        >>> src = "*IDN?"
        print(string_to_hex(src)) # [0x2A, 0x49, 0x44, 0x4E, 0x3F]
    '''
    assert isinstance(src, str)

    return [_ for _ in src.encode('utf-8')]


def string_to_byte(src: str) -> bytes:
    '''Transfer string data to bytes.

    Args:
        src (str): source string.

    Returns:
        bytes: bytes data.

    Examples:
        >>> src = "*IDN?"
        print(string_to_byte(src)) # b"*IDN?"
    '''
    assert isinstance(src, str)

    return src.encode('utf-8')


def hex_to_string(src: list) -> str:
    '''Transfer hex list to string data.

    Args:
        src (list): source list.

    Returns:
        str: string data.

    Examples:
        >>> src = [0x2A, 0x49, 0x44, 0x4E, 0x3F]
        print(hex_to_string(src)) # "*IDN?"
    '''
    assert isinstance(src, list)

    data = b''
    for _ in src:
        data += _.to_bytes(1, 'big')
    return data.decode('utf-8')


def byte_to_string(src: bytes) -> str:
    '''Transfer byte data to string data.

    Args:
        src (bytes): bytes data.

    Returns:
        str: string data.

    Examples:
        >>> src = b'*IDN'
        print(byte_to_string(src)) # "*IDN?"
    '''
    assert isinstance(src, bytes)

    return src.decode('utf-8')


def _round(number: Tuple[int, float], ndigits: int, rounding=ROUND_HALF_UP) -> float:
    assert isinstance(number, (int, float))
    assert isinstance(ndigits, int)

    _prec = len('{}'.format(number).split('.')[0]) + ndigits
    return float(Context(prec=_prec, rounding=rounding).create_decimal('{}'.format(number)))


def rounding(number: Tuple[int, float], ndigits: int = 0) -> float:
    '''
    Rounding a number to a given precision in decimal digits.

    The return value is an integer if ndigits is omitted or None.  Otherwise
    the return value has the same type as the number.  ndigits may be negative.

    Args:
        number (Tuple[int, float]): number.
        ndigits (int): ndigits may be negative.

    Returns:
        float: number.

    Examples:
        >>> number = 5.123456
        print(round(number, 4)) # 5.1235
    '''
    return _round(number, ndigits)


def round_up(number: Tuple[int, float], ndigits: int = 0) -> float:
    '''
    Rounding up a number to a given precision in decimal digits.

    The return value is an integer if ndigits is omitted or None.  Otherwise
    the return value has the same type as the number.  ndigits may be negative.

    Args:
        number (Tuple[int, float]): number.
        ndigits (int): ndigits may be negative.

    Returns:
        float: number.

    Examples:
        >>> number = 5.123456
        print(round(number, 4)) # 5.1235
    '''
    return _round(number, ndigits, ROUND_UP)


def round_down(number: Tuple[int, float], ndigits: int = 0) -> float:
    '''
    Rounding up a number to a given precision in decimal digits.

    The return value is an integer if ndigits is omitted or None.  Otherwise
    the return value has the same type as the number.  ndigits may be negative.

    Args:
        number (Tuple[int, float]): number.
        ndigits (int): ndigits may be negative.

    Returns:
        float: number.

    Examples:
        >>> number = 5.123456
        print(round(number, 4)) # 5.1234
    '''
    return _round(number, ndigits, ROUND_DOWN)
