
from base64 import encode
import os
import ast
import paramiko
import hashlib
from typing import Any, Tuple
from runtime import Runtime
import json
from re import compile as regexpCompile, IGNORECASE
from docs2md import Docs2Json
from .ssh import SSH


class _RPCDef:
    LOCALE_CONTROL_BIT = 'controlbit.json'
    END_SWITCH = {
        ' ': ' ',
        '"': '"',
        "'": "'",
        '[': ']',
        '{': '}',
        '(': ')'
    }


class CommandException(Exception):
    pass


class _RPC():
    def __init__(self, ipaddr: str, port: int) -> None:
        self._config_file_md5 = None
        self._link = None
        self.ipaddr = ipaddr
        self.port = port
        self._opened = False
        self._timeout = None
        self._name = None

    @property
    def is_open(self):
        return self._opened

    @property
    def timeout(self) -> Tuple[int, float]:
        return self._timeout

    @timeout.setter
    def timeout(self, timeout: Tuple[int, float]) -> None:
        if timeout:
            assert isinstance(timeout, (int, float))
        self._timeout = timeout


class RPCBase():
    """TinyRPC is support for setup rpc connect.

    Args:
        partno (str): Device PartNo.
        CRLF (str): Reserve.
        setting (str): DDP connect setting. Format is "ip:port".

    Attributes:
        ipaddr (str): DDP connect ip.
        port (int): DDP connect port.

    Examples:
        >>> rpc = TinyRPC(partno="Xavier-001-007A", setting="169.254.1.32:7801')
    """

    def __init__(self, partno: str = "", setting: str = "", CRLF: str = "") -> None:
        __ipaddr_port_re = regexpCompile(r'^((2[0-4]\d|25[0-5]|[1-9]?\d|1\d{2})\.){3}(2[0-4]\d|25[0-5]|[01]?\d\d?)'
                                         r'(:(6553[0-5]|655[0-2][0-9]|65[0-4]\d{2}|6[0-4]\d{3}|[1-5]\d{4}|[1-9]\d{0,3}))?$', IGNORECASE)
        if not __ipaddr_port_re.match(setting):
            raise ValueError('Setting invalid: {}'.format(setting))
        _parameters = setting.strip().split(':')
        self.ipaddr = _parameters[0]
        self.port = None
        if len(_parameters) > 1:
            self.port = int(_parameters[1])
        self._attrs = []
        self._duts = {}
        self.link = None
        self._timeout = None
        self.version = None
        self.docs_en = []
        self.docs_zh = []
        self._name = None
        self._config_path = None
        self.complete = False
        self._local_controlbit = None
        self._config_file_md5 = None
        self._default = None
        self.duts = {}

    def call(self, command: str) -> Any:
        pass

    def _delete_proxy(self):
        for attr in self._attrs:
            if hasattr(self, attr):
                delattr(self, attr)
        self._attrs = []

    def _get_current_doc(self):
        name = None
        for dut in self.duts:
            if self.link.port == self.duts[dut]:
                name = dut
                break
        if not name:
            return []

        for doc in self.docs_en:
            if doc['name'] == name:
                return doc['classes']

        return []

    def _proxy(self):
        # 删除之前添加的属性
        self._delete_proxy()
        methods = self._get_all_methods(self.link)
        for prefix in methods:
            class_name = prefix
            if prefix.endswith('.'):
                class_name = prefix[:-1]
            proxy = self._rpc_proxy(self.link, class_name)
            proxy.proxy_attributes = []
            setattr(self, class_name, proxy)
            self._attrs.append(class_name)
            for sub in methods[prefix]:
                proxy.proxy_attributes.append(sub['name'])

    def _fill_func_docs(self, func):
        func_docs_en = {}
        func_docs_en['name'] = func['name']
        func_docs_zh = {}
        func_docs_zh['name'] = func['name']
        desc, args, returns, example = Docs2Json.get_function_desc(func)
        if desc:
            func_docs_en['desc'] = desc
            func_docs_zh['desc'] = desc
        for arg in args:
            arg_list = arg.strip().strip(':').split(':')
            if len(arg_list) > 1:
                if 'args' not in func_docs_en:
                    func_docs_en['args'] = []
                func_docs_en['args'].append(
                    {'name': arg_list[0], 'desc': arg_list[1]})
            else:
                if 'desc' not in func_docs_en:
                    func_docs_en['desc'] = ''
                else:
                    func_docs_en['desc'] += '\n'
                func_docs_en['desc'] += arg
        if example:
            func_docs_en['examples'] = example
            func_docs_zh['examples'] = example
        if returns:
            func_docs_en['returns'] = {"desc": returns}
            func_docs_zh['returns'] = {"desc": returns}
        return func_docs_en, func_docs_zh

    def _docstring_to_json(self, server, docs):
        docs_en = {}
        docs_zh = {}
        docs_en['name'] = server
        docs_zh['name'] = server
        for prefix in docs:
            if not prefix:
                continue
            if 'classes' not in docs_en:
                docs_en['classes'] = []
            if 'classes' not in docs_zh:
                docs_zh['classes'] = []
            class_docs_en = {}
            class_docs_zh = {}
            if prefix.endswith('.'):
                class_docs_en['name'] = prefix[:-1]
                class_docs_zh['name'] = prefix[:-1]
            else:
                class_docs_en['name'] = prefix
                class_docs_zh['name'] = prefix
            class_docs_en['methods'] = []
            class_docs_zh['methods'] = []
            for func in docs[prefix]:
                func_docs_en, func_docs_zh = self._fill_func_docs(func)
                class_docs_en['methods'].append(func_docs_en)
                class_docs_zh['methods'].append(func_docs_zh)
            docs_en['classes'].append(class_docs_en)
            docs_zh['classes'].append(class_docs_zh)
        for sub_docs in self.docs_en:
            if sub_docs['name'] == server:
                self.docs_en.remove(sub_docs)
                break
        for sub_docs in self.docs_zh:
            if sub_docs['name'] == server:
                self.docs_zh.remove(sub_docs)
                break
        self.docs_en.append(docs_en)
        self.docs_zh.append(docs_zh)

    def _get_name(self):
        objs = Runtime.get_objs(Runtime.get_channel())
        if not objs:
            return

        for obj in objs:
            if objs[obj]["object"] == self:
                self._name = obj
                break

    def _init_config_path(self):
        self._get_name()
        self._config_path = "{0}{1}config{1}[{2}]{3}".format(
            Runtime.get_project_path(), os.sep, Runtime.get_channel(), self._name)
        self._local_controlbit = "{0}{1}{2}".format(
            self._config_path, os.sep, _RPCDef.LOCALE_CONTROL_BIT)

    def _initialization(self):
        self._get_name()
        self._config_path = "{0}{1}config{1}[{2}]{3}".format(
            Runtime.get_project_path(), os.sep, Runtime.get_channel(), self._name)
        self._local_controlbit = "{0}{1}{2}".format(
            self._config_path, os.sep, _RPCDef.LOCALE_CONTROL_BIT)
        if not os.path.exists(self._config_path):
            os.makedirs(self._config_path)
        self.checksum_file = "{0}{1}{2}".format(
            self._config_path, os.sep, self._checksum_file)
        self.en_file = "{0}{1}{2}".format(
            self._config_path, os.sep, self._docs_en_name)
        self.zh_file = "{0}{1}{2}".format(
            self._config_path, os.sep, self._docs_zh_name)
        if os.path.isfile(self.checksum_file):
            with open(self.checksum_file, 'r', encoding='utf-8') as file:
                checksum = json.load(file)
            with open(self.en_file, 'r', encoding='utf-8') as file:
                self.docs_en = json.load(file)
            with open(self.zh_file, 'r', encoding='utf-8') as file:
                self.docs_zh = json.load(file)
            data_en = json.dumps(self.docs_en, indent=4, ensure_ascii=False)
            data_zh = json.dumps(self.docs_zh, indent=4, ensure_ascii=False)
            en_md5 = hashlib.md5(data_en.encode('utf-8')).hexdigest()
            zh_md5 = hashlib.md5(data_zh.encode('utf-8')).hexdigest()
            self.complete = (
                en_md5 == checksum[self._docs_en_name] and zh_md5 == checksum[self._docs_en_name])
            self.version = checksum['version']
        else:
            self.complete = False

    def _connect_default(self):
        if self.port:
            self._default = self._rpc_client(
                self.ipaddr, self.port)
        else:
            self._default = self._rpc_client(
                self.ipaddr, self._mgnt_port)
        self.link = self._default
        self._default.open()

    def open(self) -> None:
        self._initialization()
        self._connect_default()
        self._get_xavier_duts()

    def _all_servers_docstring(self):
        link = self._rpc_client(self.ipaddr, self._mgnt_port)
        link.open()
        version = self._get_version(link)
        link.close()
        for dut in self.duts:
            if self.version != version or not self.complete:
                link = self._rpc_client(self.ipaddr, self.duts[dut])
                link.open()
                methods = self._get_all_methods(link)
                self._docstring_to_json(dut, methods)

        if self.version != version or not self.complete:
            data_en = json.dumps(self.docs_en, indent=4, ensure_ascii=False)
            with open(self.en_file, 'w+', encoding='utf-8') as file:
                file.write(data_en)
                file.flush()
            data_zh = json.dumps(self.docs_en, indent=4, ensure_ascii=False)
            with open(self.zh_file, 'w+', encoding='utf-8') as file:
                file.write(data_zh)
                file.flush()

            checksum = {}
            checksum[self._docs_en_name] = hashlib.md5(
                data_en.encode('utf-8')).hexdigest()
            checksum[self._docs_zh_name] = hashlib.md5(
                data_zh.encode('utf-8')).hexdigest()
            checksum['version'] = version
            with open(self.checksum_file, 'w+', encoding='utf-8') as file:
                json.dump(checksum, file, indent=4, ensure_ascii=False)
        self.version = version

    def _get_xavier_info(self):
        self._all_servers_docstring()
        self._proxy()

    def close(self) -> None:
        self.link.close()
        self.duts = {}
        self.version = None
        self.docs_en = []
        self.docs_zh = []
        self.complete = False
        self.link = None
        self._timeout = None
        self._delete_proxy()

    def _get_xavier_duts(self):
        data = self._get_xavier_file(self._xavier_profile)
        if not data:
            raise ConnectionError("Could not get xavier channel infomation!")
        profile = json.loads(data)
        for dut in profile['duts']:
            self.duts[dut] = profile['duts'][dut]['port']
        self.duts[self._mgnt_server] = self._mgnt_port

    def reopen(self) -> None:
        if not self._name:
            self._initialization()
        if not self.duts:
            self._get_xavier_duts()
        if not self.link:
            self.link = self._rpc_client(
                self.ipaddr, self.port if self.port else self._mgnt_port)
        try:
            self.link.close()
        except Exception:
            pass
        self.link.open()
        self._proxy()

    @property
    def is_open(self):
        if self.link:
            return self.link._opened
        return False

    @property
    def timeout(self):
        return self._timeout

    @timeout.setter
    def timeout(self, timeout) -> None:
        if timeout:
            assert isinstance(timeout, (int, float))
        self._timeout = timeout

    def _search_default_dut(self):
        if self.port:
            for dut in self.duts:
                if self.duts[dut] == self.port:
                    return dut
        return self._mgnt_server

    def switch(self, server=None):
        if not self.duts:
            self._get_xavier_duts()
        if not server:
            server = self._search_default_dut()
        assert server in self.duts
        try:
            self.link.close()
        except Exception:
            pass
        if not self.link or self.link.port != self.duts[server]:
            self.link = self._rpc_client(self.ipaddr, self.duts[server])
        self.link.open()
        self._proxy()

    def _decode_command(self, value):
        args = []
        start = None
        same_cnt = 0
        arg = ''

        for _value in value:
            if not start:
                arg += _value
                same_cnt += 1
                if _value not in _RPCDef.END_SWITCH:
                    start = ' '
                else:
                    start = _value
            else:
                if _value in _RPCDef.END_SWITCH:
                    # 字符在 _RPCDef.END_SWITCH 定义中，并且以解析内容不为空字符串
                    if not arg.lstrip(',').strip():
                        start = _value
                        arg = _value
                    elif _value == start:
                        arg += _value
                        if _value in ['"', "'"]:
                            same_cnt -= 1
                        else:
                            same_cnt += 1
                    else:
                        arg += _value
                elif _value == _RPCDef.END_SWITCH[start]:
                    arg += _value
                    same_cnt -= 1
                elif _value == ',':
                    if same_cnt != 0 and start != ' ':
                        arg += _value
                    elif start == ' ':
                        same_cnt -= 1
                else:
                    arg += _value
            if same_cnt == 0:
                arg = arg.strip().strip(',').strip()
                if arg:
                    args.append(ast.literal_eval(arg.strip()))
                start = None
                arg = ''
        if arg.strip() and start:
            arg = arg.strip().strip(',').strip()
            if arg:
                args.append(ast.literal_eval(arg.strip()))
        return args

    def _get_file_md5(self, locale_config):
        '''
        get file md5 value

        Args:
            None

        Return:
            string, file md5 string
        '''
        md5_methods = hashlib.md5()
        with open(locale_config, 'r', encoding='utf-8') as config_file_point:
            md5_methods.update(config_file_point.read().encode('utf-8'))
        new_md5 = md5_methods.hexdigest()

        return new_md5

    def _get_xavier_md5(self, xavier_config):
        if self._config_file_md5:
            return self._config_file_md5
        _ssh = SSH("", "{}@{}:{} {}".format(self._user, self.ipaddr, 22, self._passwd), "\n")
        _ssh.open()
        _ssh.timeout = 1
        try:
            _stdout = _ssh.exec(
                "md5sum {}".format(xavier_config))
        except RuntimeError:
            return None

        if _stdout and xavier_config in _stdout:
            self._config_file_md5 = _stdout.split(" ")[0]
            _ssh.close()
            return self._config_file_md5

        _ssh.close()
        return None

    def _put_json_to_xavier(self, local_path, remote_path):
        '''
        put control_bit.json file to xavier

        Args:
            local_path: string, local control_bit.json path, eg:'config/control_bit.json'
            remote_path: string, xavier control_bit.json path, eg:'/opt/seeing/app/ee/profile/control_bit.json'

        Return:
            None
        '''
        assert isinstance(local_path, str)
        assert isinstance(remote_path, str)

        _tran = paramiko.Transport((self.ipaddr, 22))
        _tran.connect(username=self._user,
                      password=self._passwd)
        sftp = paramiko.SFTPClient.from_transport(_tran)
        sftp.put(local_path, remote_path)
        sftp.close()
        _tran.close()

    def relay_switch(self, test_point: list) -> bool:
        '''
        Switch the state of the bit according to the network label defined in the control_bit file

        Args:
            test_point_name:   string,     defined by control bit file.

        Returns:
            boolen:     True: excute command success.
                        False: excute command fail.

        Examples:
            >>> ddp = TinyDDP(partno="Xavier-001-007A", setting="169.254.1.32:7801')
            print(ddp.relay_switch(['POWER_ON']))
        '''

        assert isinstance(test_point, list)

        local_md5 = self._get_file_md5(self._local_controlbit)
        xavier_md5 = self._get_xavier_md5(self._xavier_controlbit)
        if local_md5 != xavier_md5:
            self._put_json_to_xavier(
                self._local_controlbit, self._xavier_controlbit)
            self.call("io.load_relay_map('{}')".format(
                self._xavier_controlbit))
            self._config_file_md5 = local_md5
        return self.call("io.relay_switch({}, '{}')".format(test_point, self._xavier_controlbit))

    def _get_xavier_file(self, profile):
        _ssh = SSH("", "{}@{}:{} {}".format(self._user, self.ipaddr, 22, self._passwd), "\n")
        _ssh.open()
        _ssh.timeout = 5
        _stdout = _ssh.exec(
            "cat {}".format(profile))
        _ssh.close()
        return _stdout
