import re
import time
from typing import Tuple
import paramiko
import socket


class SSH():
    """SSH is support for setup a ssh client connet to server.

    Args:
        partno (str): Device PartNo.
        CRLF (str): SSH Message endle, also you can reserve blank and
                    send command with endle every time.
        setting (str): SSH clinet connect setting. Format is 'user@ip:port,password'.

    Attributes:
        ipaddr (str): SSH connect ip.
        port (int): SSH connect port.
        user (str): SSH login user name.
        password (str): SSH login password.

    Examples:
        >>> ssh = SSH(setting='root@169.254.1.32:22 123456', CRLF='\\r\\n')
    """

    def __init__(self, partno: str = "", setting: str = "", CRLF: str = "") -> None:
        self.partno = partno
        self.ipaddr = None
        self.port = None
        self.user = None
        self.password = None
        if not re.match(r'.*@((2[0-4]\d|25[0-5]|[1-9]?\d|1\d{2})\.){3}(2[0-4]\d|25[0-5]|[01]?\d\d?):'
                        r'(6553[0-5]|655[0-2][0-9]|65[0-4]\d{2}|6[0-4]\d{3}|[1-5]\d{4}|[1-9]\d{0,3})\ (.*)', setting):
            raise ValueError('Setting invalid: {}'.format(setting))
        _parameters = setting.strip().split('@')
        self.user = _parameters[0]
        _parameters = _parameters[1].strip().split(':')
        self.ipaddr = _parameters[0]
        _parameters = _parameters[1].strip().split(' ')
        self.port = int(_parameters[0])
        self.password = _parameters[1]
        self._opened = False
        self._tran = None
        self._link = None
        self.crlf = CRLF
        self._timeout = None

    def open(self):
        """
        """
        if not self._tran:
            self._tran = paramiko.Transport((self.ipaddr, self.port))
            self._tran.connect(username=self.user, password=self.password)
        if not self._link:
            self._link = paramiko.SSHClient()
        self._link._transport = self._tran
        self._opened = True

    def close(self):
        """
        """
        if self._link:
            self._link.close()
        if self._tran:
            self._tran.close()
        self._tran = None
        self._link = None
        self._opened = False

    def reopen(self):
        if self._link:
            self._link.close()
        if self._tran:
            self._tran.close()
        self._tran = None
        self._link = None
        self.open()
        self._opened = True

    @property
    def is_open(self):
        return self._opened

    @property
    def timeout(self) -> Tuple[int, float]:
        return self._timeout

    @timeout.setter
    def timeout(self, timeout: Tuple[int, float]) -> None:
        if timeout:
            assert isinstance(timeout, (int, float))
        self._timeout = timeout

    def _read(self, pip, timeout=None) -> str:
        lines = ""
        start = time.time()
        while True:
            try:
                data = pip.readline()
                if data:
                    lines += data
                else:
                    break
            except socket.timeout:
                pass

            if timeout and timeout < time.time() - start:
                raise TimeoutError('Read pipline timeout!')

        return lines

    def exec(self, command: str) -> str:
        """Execute command via ssh.

        Args:
            command (str): ssh command.

        Returns:
            str: command execute result.

        Examples:
            >>> ssh = SSH(setting='root@169.254.1.32:22 123456', CRLF='\\r\\n')
            print(ssh.exec('ls /root -l'))
        """
        assert isinstance(command, str)
        start_time = time.time()
        try:
            _, _stdout, _stderr = self._link.exec_command(
                '{}{}'.format(command, self.crlf), timeout=self._timeout)
        except paramiko.ssh_exception.SSHException as _except:
            self.close()
            raise _except

        try:
        # 读取标准输出
            lines = self._read(
                _stdout, None if not self._timeout else self._timeout - (time.time() - start_time))
            if lines:
                return lines
            # 标准输出中没有数据，则认为可能出错了，读取标准错误的数据
            lines = self._read(_stderr, None if not self._timeout else self._timeout - (time.time() - start_time))
            if lines:
                # 读到出错信息时，抛出报错信息
                raise RuntimeError(lines)
        except TimeoutError:
            raise TimeoutError('Execute timeout!')

        return lines
