
import re
import os
from typing import Any
from ._rpc import _RPC
from ._rpc import RPCBase
from ._rpc import CommandException
from tinyrpcx.rpc_client import RPCClientWrapper


class TinyRPCXDef:
    XAVIER_CONTROL_BIT = '/mix/addon/config/controlbit.json'
    XAVIER_PROFILE = '/mix/addon/config/hw_profile.json'
    DEFAULT_TIMEOUT = 0xFFFFFFF
    USER = 'root'
    PASSWD = '123456'
    DOCS_ZH_NAME = 'docstring_zh.json'
    DOCS_EN_NAME = 'docstring_en.json'
    SYS_DOCS_ZH_FILE = os.path.abspath(
        "{0}{1}..{1}config{1}{2}".format(os.getcwd(), os.sep, DOCS_ZH_NAME))
    SYS_DOCS_EN_FILE = os.path.abspath(
        "{0}{1}..{1}config{1}{2}".format(os.getcwd(), os.sep, DOCS_EN_NAME))
    VERSION = '/mix/version.json'

    MGNT_SERVER = 'mgnt'
    MGNT_PORT = 7800
    CHECKSUM_FILE = 'checksum.json'


class _TinyRPCXClient(_RPC):
    """TinyRPC is support for setup rpc connect.

    Args:
        partno (str): Device PartNo.
        CRLF (str): Reserve.
        setting (str): DDP connect setting. Format is "ip:port".

    Attributes:
        ipaddr (str): DDP connect ip.
        port (int): DDP connect port.

    Examples:
        >>> rpc = TinyRPC(partno="Xavier-001-007A", setting="169.254.1.32:7801')
    """

    def __init__(self, ipaddr: str, port: int) -> None:
        _RPC.__init__(self, ipaddr, port)
        self._ssh = None
        self._sftp = None
        self._tran = None

    def open(self) -> None:
        """
        """
        endpoint = {"requester": "tcp://{}:{}".format(self.ipaddr, self.port),
                    "receiver": "tcp://{}:{}".format(self.ipaddr, self.port + 10000)}
        self._link = RPCClientWrapper(endpoint).rpc_client
        self._opened = True
        setattr(self, "call", self._link.call)

    def close(self) -> None:
        """
        """
        try:
            self._link.stop()
        except Exception:
            pass
        if hasattr(self, 'call'):
            delattr(self, 'call')
        self._link = None
        self._opened = False

    def reopen(self) -> None:
        try:
            self._link.stop()
        except Exception:
            pass
        self.open()


class TinyRPCX(RPCBase):
    """TinyRPC is support for setup rpc connect.

    Args:
        partno (str): Device PartNo.
        CRLF (str): Reserve.
        setting (str): DDP connect setting. Format is "ip:port".

    Attributes:
        ipaddr (str): DDP connect ip.
        port (int): DDP connect port.

    Examples:
        >>> rpc = TinyRPC(partno="Xavier-001-007A", setting="169.254.1.32:7801')
    """

    def __init__(self, partno: str = "", setting: str = "", CRLF: str = "") -> None:
        RPCBase.__init__(self, partno, setting, CRLF)
        self._xavier_controlbit = TinyRPCXDef.XAVIER_CONTROL_BIT
        self._xavier_profile = TinyRPCXDef.XAVIER_PROFILE
        self._user = TinyRPCXDef.USER
        self._passwd = TinyRPCXDef.PASSWD
        self._docs_zh_name = TinyRPCXDef.DOCS_ZH_NAME
        self._docs_en_name = TinyRPCXDef.DOCS_EN_NAME
        self._sys_docs_en_file = TinyRPCXDef.SYS_DOCS_EN_FILE
        self._sys_docs_zh_file = TinyRPCXDef.SYS_DOCS_ZH_FILE
        self._version = TinyRPCXDef.VERSION
        self._mgnt_server = TinyRPCXDef.MGNT_SERVER
        self._mgnt_port = TinyRPCXDef.MGNT_PORT
        self._checksum_file = TinyRPCXDef.CHECKSUM_FILE
        self._rpc_proxy = TinyRPCXProxy
        self._rpc_client = _TinyRPCXClient

    def call(self, command: str) -> Any:
        assert isinstance(command, str)

        group = re.match(r'.*\(.*\)$', command.strip())
        if not group:
            raise CommandException(
                'Invalid command: {}'.format(command))
        group = re.split(r'\(|\)', command)
        arg = self._decode_command(group[1])
        if self.link._timeout:
            kwargs = {'timeout_ms': self.link._timeout * 1000}
        else:
            kwargs = {'timeout_ms': TinyRPCXDef.DEFAULT_TIMEOUT}
        return self.link.call(group[0], *arg, **kwargs)

    def _get_version(self, link):
        kwargs = {'timeout_ms': 500}
        return link.call(
            "xavier.fw_version", *[], **kwargs)

    def _get_all_methods(self, link):
        kwargs = {'timeout_ms': 1000}
        return link.call(
            "server.all_methods", *[], **kwargs)


class TinyRPCXProxy():
    def __init__(self, link, prefix) -> None:
        self._link = link
        self._prefix = prefix

    def __getattr__(self, __name: str) -> Any:
        if self._link._timeout:
            kwargs = {'timeout_ms': self._link._timeout * 1000}
        else:
            kwargs = {'timeout_ms': TinyRPCXDef.DEFAULT_TIMEOUT}
        func = lambda args=[]: self._link.call(
            "{}.{}".format(self._prefix, __name), *args, **kwargs)
        func.__name__ = "{}.{}".format(self._prefix, __name)
        return func
