# 插件开发指引

Archon 插件是 Archon 平台某一类仪器或者功能提供统一的接口，方便测试用例工程师复用并简化测试用例的编写。

## 开发新插件指导
<font color=red>注: 插件涉及的python代码需要遵循 python3.6 以上版本的特性。</font>

### 插件代码目录

```dir
AAAPlugin_dir
    |----AAAPlugin           # 插件代码目录
    |       |----__init__.py # 插件对外接口
    |       |----AA1.py      # 插件代码
    |       |----AA2.py      # 插件代码
    |       |----AA1_cn.md   # AA1插件的中文文档
    |       |----AA1_en.md   # AA1插件的英文文档
    |       |----AA2_cn.md
    |       |----AA2_en.md
    |----ChangeLists.md      # 插件的修改记录
    |----setup.py            # 插件打包脚本
    |----tests               # 插件单元测试目录
            |----__init__.py
            |----test_AA1.py
            |----test_AA2.py
```

从上述目录中可以看出主要包含的文件目录有：插件代码目录(`AAAPlugin`)，单元测试目录(tests)，插件修改记录文件(ChangeLists.md)，插件打包脚本(setup.py)。

![](resource/how_to_write_plugin/new_project_dir.png)

### 添加插件代码

1. 插件代码文件： 指的是插件的功能性代码，可以是python文件、c语言文件、编译好的lib文件等。

2. python 代码文件建议与插件类名保持一致，但是名字为小写。

![](resource/how_to_write_plugin/file_and_class_name.png)

3. 在python插件代码目录中的`__init__.py`文件中导入对外发布的模块和独立接口，添加`__all__`、`__type__`、`__version__`等字段。

- `__all__`: 对外发布的独立接口或者模块名称。
- `__type__`: 插件的类型，现有类型有：dmm、audio、scope、dcpower、pwrmeter、specan、switch
- `__version__`: 插件版本号。

![](resource/how_to_write_plugin/init_code.png)

4. 禁止插件函数使用多重连接符调用，所有需要使用的接口都应该只有一层调用。

错误的代码例子：

```python
# 代码
class Measure(object):
    def voltage(self):
        # measure voltage code
        return volt

    def current(self):
        # measure voltage code
        return current

    def power(self):
        # measure voltage code
        return power


class ETM605P(PowerBase):
    '''
    ETM605P plugin
    '''
    compatible = ['ETM605P']
    def __init__(self, instrument_name):
        super(ETM605P, self).__init__(instrument_name)
        self._measure = Measure(self.modbus, self._device_addr, self.device_info)

    def measure(self):
        return self._measure

# 用户使用
ETM605P('ETM605P').measure().voltage()
```

正确的代码例子：
```python
# 代码
class ETM605P(PowerBase):
    '''
    ETM605P plugin
    '''
    compatible = ['ETM605P']
    def __init__(self, instrument_name):
        super(ETM605P, self).__init__(instrument_name)

    def voltage(self):
        # measure voltage code
        return volt

    def current(self):
        # measure voltage code
        return current

    def power(self):
        # measure voltage code
        return power

# 用户使用
ETM605P('ETM605P').voltage()
```

5. 配置类接口建议统一使用`set_`作为接口前缀。

6. 功能类接口的命名方式建议直接使用功能的名称作为接口名称，而不是表述不清的单词或者是自定义的简写。

7. 使用pylint指令运行静态检查，统一pylint标准参照[pep-8](https://peps.python.org/pep-0008/)。

### 仪器插件属性

1. 仪器类插件中需要`compiltable`字段中添加插件支持的硬件信息。
    - 如果板卡存在多种硬件代号，需要全部填写。
    - 如果插件支持多款硬件，也需要把硬件型号添加上。

![](resource/how_to_write_plugin/compiltable.png)

2. 仪器类插件中需要有`attachment`字段，包含vendor和connect信息。
    - vendor 填写的是仪器的供应商名称。
    - connect 填写的是仪器的基本连接信息。字段信息是字典，每个元素的填写的规则是`类型`或者`类型://默认连接参数`，其中现在支持的`类型`包含`Xavier`(通过xavier控制的仪器)、`visa`(通过visa协议控制的仪器，连接方式可以是UART、TCPIP、GPIB、USB)、`uart`(用串口发命令控制的仪器)、`tcp`(TCP/IP协议控制的仪器)。而`默认连接参数`的填写规则是插件初始化函数的`参数名称=默认值`，每个参数都用`&`号进行连接。

connet字段是`类型`示例：

![](resource/how_to_write_plugin/connect_xavier.png)

connet字段是`类型://默认连接参数`示例：

![](resource/how_to_write_plugin/connect_visa.png)

3. 仪器相关的选项必须写在插件代码中，包括但不限于：通信时的结束符。

![](resource/how_to_write_plugin/CRLF.png)

### 单元测试

1. 单元测试的文件命名必须是插件代码文件名称加上`test_`前缀。

![](resource/how_to_write_plugin/unitest_name.png)

2. 导入模块并使用mock编写测试代码。

3. 使用`pytest -vv --cov-config .coveragerc --pep8 --cov=./ --cov-report=term --cov-report=html:htmlcov` 指令运行单元测试，要求覆盖率达到90%以上并无错误才能算通过。

### 中英文文档

<font color=red>在没有双语言文档的情况下，可以使用docstring代替markdown文档，关于docstring的规则，具体请参考[google style](https://google.github.io/styleguide/pyguide.html?showone=Comments#s3.8-comments-and-docstrings)。</font>

关于中英文文档的要求如下：

1. 用户可以根据自身所需要的语言编写中文或英文文档，Archon会根据所选择的语言显示接口提示信息；在没有任何接口文档的情况下，页面只显示接口名称与参数名称，不显示任何接口提示信息。

2. 文档的一级标题固定是`Model(型号)`，内容是对应的硬件型号，硬件型号以`*`号开头，例子如下：

    ```en
    # Model
    * DMM003003
    ```

    ```cn
    # 型号
    * DMM003003
    ```

3. 文档的二级目录有`Description(描述)`、`Vendor(厂家)`和接口定义。

    其中`Description(描述)`需要写清楚插件或者对应仪器的功能说明，`Vendor(厂家)`则是需要写清楚仪器的开发商。例子如下：

    ```en
    ## Description
    The DMM003003 is a fully functional digital multimeter.The ADC resolution is 24 bits and can be used as a high performance DMM for DC and AC measurements.Power and ground are isolated from external power input and digital IOs, with the best anti-noise capability.


    ## Vendor
    SG
    ```

    ```cn
    ## 描述
    DMM003003是一个全功能的数字万用表。ADC 分辨率为24 位，可作为高性能 DMM 进行直流和交流测量。 电源和地与外部电源输入和数字IOs 隔离，具有最佳的抗噪声能力。


    ## 厂家
    SG
    ```

4. 每个接口定义下的三级标题名称分别为`Function Description(接口描述)`、`Parameters(参数)`、`Returns(返回值)`、`Examples(例子)`，前面是英文文档的标题后面是中文文档的标题，而无论中文还是英文文档，接口定义均是代码中的实际定义。

    关于`Parameters(参数)`字段的内容，如果接口没有参数，则整个字段下面填写`* None`即可。如果存在参数的情况下，每个参数都是以`*`符号作为小标题，下面是具体的说明，一般包含的说明有`类型`、`范围`、`默认值`、`单位`和`描述`。第一个字段必须是`类型`，最后一个字段必须是`描述`。其它字段如果存在则需要填写，填写时需要遵守以下规则：
    - 类型和描述字段直接写清楚内容即可。
    - 参数范围以`[]`进行描述，如果存在多个类型的参数类型，那么`[]`内有多个对应的范围描述。
    - 默认值字段英文以`default value is`开始，中文以`默认值是`开始。
    - 单位字段英文以`unit`开始，中文以`单位：`开始。
    - 关于条件选择型参数范围描述，除了按照上述的参数描述规则外，还需要额外描述参数选择的条件下的范围，该条件描述填写在相应的参数描述下面。英文以`if XX is YY [values]`，中文以`如果XX为YY时，参数可选[values]`来描述，其中`XX`为该条件选择型参数的条件，`YY`就是该参数的对应的条件的值，`[values]`是在该条件下参数的可选范围值，按照上述参数范围描述的规则来填写。
    - 对于一些通用或者是约定俗成的英文，不需要特意翻译成中文，如`Min`、`Max`等，不需要翻译成`最小值`、`最大值`等。

    英文示例：
    ```example
    ## voltage(count)
    ###  Function Description
    Measure the voltage
    ### Parameters
    * count: int,[1 ~ 512],default 1 value is,1 mean single measure, > 1 mean multiple points measure.
    ### Returns
    float, unit V
    ### Examples
    dmm = DMM003003("dmm003003_0")
    dmm.range("6VDC")
    dmm.sampling_rate(5)
    dmm.voltage(1)
    ```

    中文示例：
    ```example
    ## voltage(count)
    ### 测量电压
    ### 接口描述
    测量电压
    ### 参数
    * 测量次数： 整型，[1 ~ 512]， 默认值是1，1代表单次测量，大于1代表多次测量取平均值
    ### 返回值
    浮点型，单位： V
    ### 例子
    dmm = DMM003003("dmm003003_0")
    dmm.range("6VDC")
    dmm.sampling_rate(5)
    dmm.voltage(1)
    ```

### ChangeList

`ChangeLists.md`需要按照以下模板来填写：

```markdown
# Plugin Change Lists

Version sequence number - A.B

---
### VA.B
- **Modifier**: smartgain
- **Date**:  2020/2/26
- **History:**
    - 1. Add DMM001001 plugin functional.
---
```

1. 由于每次提交代码都需要修改`ChangeLists.md`文件。  
2. 版本号以二级版本号规则为准，即A.B的方式，其中A为大版本号，B为小版本号，不建议更多级的版本号规则。  
3. 每次提交代码并不代表会有新的版本发布，可能是多次提交代码才发布一次版本，因此不需要每次提交都要更新`ChangeLists.md`的版本号，而是应该视情况更新。  
4. 每次提交代码理论上都应该修改`History`的内容，如果在版本号未更新的情况下，`History`的内容应该是做增量修改的，同时需要更新`Date`字段的时间。  
5. `History`的内容不能是无意义的描述，必须描述清楚代码修改的内容，新增功能建议以`Added`开始，修复的Bug建议以`Fixed`开始，其它功能的变更需要描述清楚。  
    错误的模板：  
    ```markdown
    - **History:**
        - 1. Init draft.        # 初始版本，那么请说清楚初始版本的功能。
        - 2. update md file.    # 更新md文件就应该说清楚更新的内容。
    ```

    正确的模板：  
    ```markdown
    - **History:**
        - 1. Added DMM005003 unit test, coverage is 90%.
        - 2. Fixed DMM003003 simplate tips error, from 125000 to 1250000.
    ```

### 插件打包脚本

插件的打包脚本是需要按照python中的setup tools规则来编写。

需要注意以下几点：

1. 由于开发插件时可能会存在第三方库依赖的情况，如果使用了第三方库，则需要在`setup.py`文件中填写`install_requires`字段内容，包括依赖库的名称与版本号信息，详细的填写规则说明请查看[pep 0508](https://www.python.org/dev/peps/pep-0508/#environment-markers)。

```example
install_requires = [
    "paramiko>=2.7.1",
    "six~=1.15.0",
    "toml==0.10.1",
    "wcwidth>=0.2.5",
    "zipp>=3.1.0",
]
```

2. 如果插件包需要包含非python文件，则需要在`setup.py`中添加的`include_package_data=True`，并且把文件的过滤规则填写到`manifest.in`文件中。`manifest.in`具体的规则可以参考[including-data-files](https://setuptools.readthedocs.io/en/latest/setuptools.html#including-data-files)。

`manifest.in`的例子如下：
```example
recursive-include vna *.md
```

3. `author`统一填写`SmartGiant`，`author_email`统一填写`devops@gzseeing.com`。

4. 使用`python setup.py $version bdist_wheel`指令打包成插件包，其中`$version`是指插件的版本号。

## 提交合并请求

提交合并请求时需要填写以下内容：

1. 写清楚修改的内容
2. 添加单元测试结果
3. 添加pylint检测结果

        修改内容:

        1. Item 1 desc
        2. Item 2 desc

        单元测试结果:
        ```
        ================================================== warnings summary ===================================================
        ----------- coverage: platform win32, python 3.6.8-final-0 -----------
        Name               Stmts   Miss  Cover
        --------------------------------------
        dmm\dmm001001.py      27      0    100%
        dmm\dmm003003.py      32      0    100%
        --------------------------------------
        TOTAL                 59      0    100%
        Coverage HTML written to dir htmlcov

        =========================================== 40 passed, 6 warnings in 0.61s ============================================

        ```

        静态检查:
        ```
        $ pylint --output-format=parseable --rcfile=pylint.conf dmm

        --------------------------------------------------------------------
        Your code has been rated at 10.00/10 (previous run: 10.00/10, +0.00)
        ```
